/*
 * Servocontroller 1.5
 * © Fenna Pel, fenna@pereltje.nl
 * 
 * gebaseerd op dit artikel:
 * https://learn.adafruit.com/multi-tasking-the-arduino-part-2/overview by Bill Earl
 * 
 * De code is aangepast en deels herschreven om tot 7 servo's te kunnen aansturen met instelbare snelheid en uitslag.
 * Om de aanpassingen heen is nieuwe code geschreven om de speciefieke eigenschappen van de gebruikte servo's netjes op te vangen en om functionaliteit toe te voegen.
 * 
 * voor versie 1.5 (2020-11-17) 
 *   - code verder opgeschoond
 *   - 'int' types voor 'short' vervangen waar mogelijk, voorsorteren op de opslag van de waarden in de EEPROM en zuiniger op het RAM gebruik
 *   - servo aansturing via array's ipv gewone variabelen
 *   - Een aantal hardgecodeerde constantes voor systeem variabelen vervangen
 *   - Eerste aanzet voor range programeren (placeholder subroutines aangemaakt voor lezen/schrijven EEPROM)
 *   - code terug geschreven voor seriele monitor
 *   - Een servo monitor subroutine toegevoegd voor seriele monitor
 *   - Bij elke timer interrupt (1/ms) wordt slechts 1 servo bijgesteld ipv alle 7 (geeft soepelere slag op alle servo's) Een teller houdt bij welke servo aan de beurt is
 *   - 
 *   
 * voor versie 1.4
 * 
 * voor versie 1.3 (2020-10-31)
 *   - code opgeschoond en hard gecodeerde waarden voor pinnen omgeschreven naar vooraf gedefinieerde variabelen/constanten
 *   - code klaargemaakt voor 7 servo's ipv de 2 hard gecodeerde servo's
 *   - gracefull shutdown geeft nu een knipperende power led en is in een aparte routine ondergebracht
 *   - verwijzingen en code restjes voor seriele monitor verwijderd. voor een monitor kan evt een I2C display worden ingeprogrammeerd
 *   
 */



#include <Servo.h> 
#include <Wire.h>
#include <Adafruit_ADS1015.h>

// systeem variabelen en constanten
const short _SERVORANGE_MIN = 1000;
const short _SERVORANGE_MAX = 2000;
const short _SERVORANGE = _SERVORANGE_MAX - _SERVORANGE_MIN;
const short _SERVOCENTRE = (_SERVORANGE_MIN + _SERVORANGE_MAX)/2;
short _BOARDPOWER; // wordt ook gebruikt in de sweeper class
const short _SHUTDOWN_DELAY = 10000;
const short _START_DELAY = 1000;
const short _INCREMENT = 3;
const short _REVERSE_DELAY = 3000;
short _INTERRUPT_COUNTER = 0;
const short _ADA_ADC_5V = 1695; //deze waarde is mogelijk individueel hardware afhankelijk van de gebruikte I2C ADC. Met de servo print routine is de waarde te achterhalen
const short _ARD_ADC_5V = 1023; //vast gegeven voor hoe de arduino eigen ADC's werken. Voor code consistentie via deze systeem constante
const short _SWEEP_MIN = 1;
const short _SWEEP_MAX = 50;
const short _ADS1015_I2C_1 = 0x48; //I2C adres van de gebruikte adafruit I2C ADC

// pinverdeling
short potmeterpin[7] = {A0,A1,A2,A3,0,1,2};
short servopin[7] = {6,7,8,9,10,11,12};
short potmeterpin8 = 3;
short ledpin = 13; //powerled
short ledadr0 = 3; // deze 3 pins gaan naar 3-8 decoder voor indicatie bij welke servo de range wordt ingesteld
short ledadr1 = 4; //
short ledadr2 = 5; //
short poweroffpin = A7;
short programpin = A6;
short servoonpin = 2;

//  variabelen met hun startstanden:
short potmeter[7] = {0,0,0,0,0,0,0};;
short potmeter8 = 0;
short timeoff_Pon = 1;
short timeoff_Poff = 250;
short timeon = 250;
short ledadr = 7;
short programkeyvalue = 0;
short programkey = 0;
short programkey2 = 0;
short servospeed[7] = {_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN,_SWEEP_MIN};
short servorange[8] = {_SERVORANGE_MAX,_SERVORANGE_MAX,_SERVORANGE_MAX,_SERVORANGE_MAX,_SERVORANGE_MAX,_SERVORANGE_MAX,_SERVORANGE_MAX,0};

// klasse defenities:
class Flasher
{
  // Class Member Variables
  // These are initialized at startup
  short ledPin;      // the number of the LED pin
  long OnTime;     // milliseconds of on-time
  long OffTime;    // milliseconds of off-time
  //int updatetimeoff = Offtime;
 
  // These maintain the current state
  short ledState;                 // ledState used to set the LED
  unsigned long previousMillis;   // will store last time LED was updated
 
  // Constructor - creates a Flasher 
  // and initializes the member variables and state
  public:
  Flasher(short pin, long on, long off)
  {
  ledPin = pin;
  pinMode(ledPin, OUTPUT);     
    
  OnTime = on;
  OffTime = off;
  
  ledState = LOW; 
  previousMillis = 0;
  }

 void refresh( int updatetimeoff)
  {
    OffTime=updatetimeoff;
  }
  void Update(unsigned long currentMillis, int updatetimeoff)
  {
    if((ledState == HIGH) && (currentMillis - previousMillis >= OnTime))
    {
      ledState = LOW;  // Turn it off
      previousMillis = currentMillis;  // Remember the time
      digitalWrite(ledPin, ledState);  // Update the actual LED
    }
    else if ((ledState == LOW) && (currentMillis - previousMillis >= OffTime))
   // else if ((ledState == LOW) && (currentMillis - previousMillis >= opdatetimeoff))
    {
      ledState = HIGH;  // turn it on
      previousMillis = currentMillis;   // Remember the time
      digitalWrite(ledPin, ledState);   // Update the actual LED
    }
  }
};

class Sweeper
{
  Servo servo;              // the servo
  int  updateInterval;      // interval between updates
  
  volatile short pos = _SERVOCENTRE;                  // current servo position 
  volatile short pos_exec ;
  volatile unsigned long lastUpdate; // last update of position
  volatile short increment;            // increment to move for each interval

public: 
  Sweeper(short interval)
  {
    updateInterval = interval;
    increment = _INCREMENT;
  }
  
  void Attach(short pin)
  {
    servo.attach(pin);
  }
  
  void Detach()
  {
    servo.detach();
  }
  
  void refresh(short newinterval, short updateservorange)
  {
    updateInterval=newinterval;
    
  }
  
  void Update(unsigned long currentMillis, int updateservorange)
  {
    if((currentMillis - lastUpdate) > updateInterval)  // time to update
    {
      if (_BOARDPOWER == LOW && pos == _SERVOCENTRE)
      {
        updateservorange = 0;
      }
      lastUpdate = currentMillis;
      pos += increment;
      servo.writeMicroseconds(pos);
     
      if ((pos >= _SERVOCENTRE+(updateservorange/2)) || (pos <= _SERVOCENTRE-updateservorange/2)) // end of sweep
       {
        //delay the reverse and hold position:
        // reverse direction
        increment = -increment;
      }
    }
    // voorwaardelijke loop voor opstart met servo's in centrale positie
    if (updateInterval == 0) 
    {
      servo.writeMicroseconds(_SERVOCENTRE);
    }
  }
};

Adafruit_ADS1015 ads1015(_ADS1015_I2C_1); 
Flasher led1(ledpin, timeon, timeoff_Pon);
Sweeper sweeper[7](0);

// subroutines:
void adjustrange()
{
  ledadr =ledadr+1;
}

void setrange()
{
  ledadr = 7;
}

void servostart()
{
  // placeholder voor lezen EEPROM voor start positie en range voor elke servo
  delay(_START_DELAY);
}

void servostop()
{
  // placeholder voor schrijven naar EEPROM van stop positie en range voor elke servo
  delay (_SHUTDOWN_DELAY);
}

void gracefullshutdown()
{
  led1.refresh(timeoff_Poff);
  digitalWrite(ledadr0, HIGH);
  digitalWrite(ledadr1, HIGH);
  digitalWrite(ledadr2, HIGH);
  servostop();
  digitalWrite (ledpin, LOW);
  digitalWrite(servoonpin, LOW);
}

void printservo(short servonr)
{
  Serial.print ("servo: ");
  Serial.print (servonr);
  Serial.print (" , potmeterstand: ");
  Serial.print (potmeter[servonr-1]);
  Serial.print (" snelheid: ");
  Serial.print (servospeed[servonr-1]);
  Serial.print (" range: ");
  Serial.println (servorange[servonr-1]);
}

// setup routine:
void setup() 
{
  Serial.begin(9600);
  for (short i = 0; i <= 6; i++)
  {
    sweeper[i].Attach(servopin[i]);
  }

  ads1015.begin();
  
  // Timer0 is already used for millis() - we'll just interrupt somewhere
  // in the middle and call the "Compare A" function below
  OCR0A = 0xAF;
  TIMSK0 |= _BV(OCIE0A);
  
//gracefull start
servostart();
pinMode(servoonpin, OUTPUT);
digitalWrite(servoonpin, HIGH);
pinMode(poweroffpin, INPUT);
pinMode(ledadr0, OUTPUT);
digitalWrite(ledadr0, HIGH);
pinMode(ledadr1, OUTPUT);
digitalWrite(ledadr1, HIGH);
pinMode(ledadr2, OUTPUT);
digitalWrite(ledadr2, HIGH);
} 

// Interrupt is called once a millisecond, 
SIGNAL(TIMER0_COMPA_vect) 
{
  unsigned long currentMillis = millis();
  sweeper[_INTERRUPT_COUNTER].Update(currentMillis,servorange[_INTERRUPT_COUNTER]);
  _INTERRUPT_COUNTER = _INTERRUPT_COUNTER +1;
  if (_INTERRUPT_COUNTER == 7) _INTERRUPT_COUNTER = 0;
  led1.Update(currentMillis,timeoff_Pon);  
} 

// hoofd programma:
void loop()
{
  for (short i = 0; i <= 3; i++) 
  {
    potmeter[i] = analogRead(potmeterpin[i]);
    servospeed[i] = map(potmeter[i], 0,_ARD_ADC_5V,_SWEEP_MIN,_SWEEP_MAX);
  }

  for (short i = 4; i <= 6; i++) 
  {
    potmeter[i] = ads1015.readADC_SingleEnded(potmeterpin[i]);
    servospeed[i] = map(potmeter[i], 0,_ADA_ADC_5V,_SWEEP_MIN,_SWEEP_MAX);
  }

  // uncomment om op de seriele monitor de tussen de haakjes vermelde servo te volgen (1 tot en met 7)
  //printservo(1); 

  potmeter8 = ads1015.readADC_SingleEnded(potmeterpin8);
  servorange[ledadr] = map(potmeter8, 0,_ADA_ADC_5V,(_SERVORANGE/10),_SERVORANGE_MAX);

  for (short i = 0; i <= 6; i++) 
  {
    sweeper[i].refresh(servospeed[i],servorange[i]);
  }

  _BOARDPOWER = analogRead (poweroffpin);
  if (_BOARDPOWER == LOW) gracefullshutdown();

  programkeyvalue = analogRead(programpin);
  programkey = map(programkeyvalue, 1,1024,0,3);
  
  digitalWrite(ledadr0, bitRead(ledadr,0));
  digitalWrite(ledadr1, bitRead(ledadr,1));
  digitalWrite(ledadr2, bitRead(ledadr,2));
  delay(500);
  programkeyvalue = analogRead(programpin);
  programkey2 = map(programkeyvalue, 1,1024,0,3);
  if (programkey == 2 && programkey2 == 2) adjustrange();
  if (programkey == 1 && programkey2 == 1) setrange();
  if (ledadr == 8) ledadr=0;
}
